/************************************************************************
*Copyright (C) 2003 Elliott Michell					*
*Elliott Mitchell takes no responsibility for failures, or damage done	*
*by this program.							*
*									*
*$Id: fullrle.c,v 1.2 2003/11/10 07:17:38 ehem Exp $			*
*************************************************************************
*An /almost/ pure RLE version. The encoding starts with a single-byte	*
*length token. If the length is between 0 and MAXLEN, what follows is	*
*length+1 raw bytes from the input stream (the +1 is used because	*
*otherwise 0 would be unused); otherwise what follows is a byte that	*
*needs to be repeated length-MAXLEN+2 times (again, MAXLEN equates to 2	*
*because that is the minimum repeat length).				*
*									*
*To make this pure RLE, *every* symbol would need to include a length	*
*byte; as you'll note this is pretty darn inefficient because most	*
*bytes/symbols are single non-repeating ones, therefore allocating a	*
*bunch of possible values to indicate a string of non-repeating bytes	*
*is a big win. FYI this is the same trivial difference between		*
*LZ77/LZSS and LZ78/LZW (the LZW patent should never of been granted	*
*because it is *trivial*).						*
*									*
*In general stacking compression algorithms works very badly, because	*
*the second will have to deal with something that looks a lot more	*
*random, worse portions that the second /could/ of done well on may be	*
*heavily broken up. Placing Huffman on top of RLE isn't that bad	*
*because RLE doesn't disturb the input that much, however Huffman can	*
*deal with the redundancies that RLE eliminates just as well as RLE	*
*does. The one exception is that layering lossless compression on top	*
*of lossy works pretty well since the lossy compression will remove the	*
*worst randomness and generate an output stream that is a lot more	*
*regular than lossless compression does.				*
************************************************************************/

#include <stdio.h>

#ifndef MAXLEN
#define MAXLEN 128
#endif

int main(int argc, char **argv, char **envp)
{
	char *func;
	char buf[MAXLEN];
	unsigned char len;
	int tok, last;

	func=*argv+strlen(*argv);
	while(func[-1]!='/'&&func>*argv) --func;

	if(strcmp("unfullrle", func)) {
		len=0;
		last=getchar();
		while((tok=getchar())!=EOF) {
			if(tok==last) { /*A run!*/
				if(len) {
					putchar(len-1);
					fwrite(buf, 1, len, stdout);
				}
				len=MAXLEN; /*equiv to 2 char run*/
				while((tok=getchar())==last&&len<255) ++len;
				putchar(len);
				putchar(last);
				len=0;
			} else {
				buf[len++]=last;
				if(len>=MAXLEN) { /*Hmm, overflow*/
					putchar(len-1);
					fwrite(buf, 1, len, stdout);
					len=0;
				}
			}
			last=tok;
		}
		if(last!=EOF) buf[len++]=last;
		if(len) {
			putchar(len-1);
			fwrite(buf, 1, len, stdout);
		}
	} else {
		while((tok=getchar())!=EOF) {
			if(tok>=MAXLEN) { /*Run*/
				len=tok-MAXLEN+3; /*Get actual length*/
				if((tok=getchar())==EOF) break;
				while(--len) putchar(tok);
			} else { /*Raw*/
				len=tok+1;
				len=fread(buf, 1, len, stdin);
				fwrite(buf, 1, len, stdout);
			}
		}
	}
	return(0);
}

